from .discover import Discover

from zeroconf import Zeroconf, ServiceBrowser, ServiceStateChange
import os, socket, logging, time

logger = logging.getLogger(os.path.splitext(os.path.basename(__file__))[0])


class DiscoverMdns(Discover):
    """zeroconf device discovery"""

    def __init__(self, scan_rate:float=5):
        super().__init__()

    def scan(self):
        # return url's of devices that are online
        self.urls = set()
        if False:
            # asyncio ... task exception error (see below)
            zc = Zeroconf()
            try:
                ServiceBrowser(zc, "_mp._tcp.local.", self)
                ServiceBrowser(zc, "_ws._tcp.local.", self)
                ServiceBrowser(zc, "_telnet._tcp.local.", self)
                # give the scanner some time to find everything ...
                time.sleep(1)
            except Exception as ex:
                print(ex)
                logger.exception(f"Unhandled exception in DiscoverMdns._scanner")
            finally:
                zc.close()
        return self.urls

    def _url(self, zeroconf, service_type, name):
        info = zeroconf.get_service_info(service_type, name)
        if info == None: return
        ip = socket.inet_ntoa(info.addresses[0])
        port = info.port
        scheme = service_type.split('.')[0][1:]
        url = f"{scheme}://{ip}:{port}"
        return url

    def add_service(self, zeroconf, service_type, name):
        self.urls.add(self._url(zeroconf, service_type, name))

    def update_service(self, zeroconf, service_type, name):
        self.add_service(zeroconf, service_type, name)

    def remove_service(self, zeroconf, service_type, name):
        pass


"""
13.07.21 17:02:20 (-0700)  jupyter  ERROR:asyncio:Task exception was never retrieved
13.07.21 17:02:20 (-0700)  jupyter  future: <Task finished name='Task-1' coro=<AsyncEngine._async_setup() done, defined at /usr/local/lib/python3.8/site-packages/zeroconf/_core.py:104> exception=OSError(9, 'Bad file descriptor')>
13.07.21 17:02:20 (-0700)  jupyter  Traceback (most recent call last):
13.07.21 17:02:20 (-0700)  jupyter    File "/usr/local/lib/python3.8/site-packages/zeroconf/_core.py", line 107, in _async_setup
13.07.21 17:02:20 (-0700)  jupyter      await self._async_create_endpoints()
13.07.21 17:02:20 (-0700)  jupyter    File "/usr/local/lib/python3.8/site-packages/zeroconf/_core.py", line 133, in _async_create_endpoints
13.07.21 17:02:20 (-0700)  jupyter      transport, protocol = await loop.create_datagram_endpoint(lambda: AsyncListener(self.zc), sock=s)
13.07.21 17:02:20 (-0700)  jupyter    File "/usr/local/lib/python3.8/asyncio/base_events.py", line 1231, in create_datagram_endpoint
13.07.21 17:02:20 (-0700)  jupyter      sock.setblocking(False)
13.07.21 17:02:20 (-0700)  jupyter  OSError: [Errno 9] Bad file descriptor
"""


from .device_registry import DeviceRegistry
from .eval import RemoteError
from .utilities import cd, redirect_stdout_stderr
from .secrets import Secrets
from .env import Env
from .device_config import DeviceConfig
from .version import __version__



from .discover import Discover
from .secrets import Secrets

from threading import Thread, Lock

import socket, threading, os, time, logging

logger = logging.getLogger(os.path.splitext(os.path.basename(__file__))[0])


class DiscoverBroadcasts(Discover):

    def __init__(self, max_age=5):
        """Start a daemon thread that collects advertisements and discards them after max_age seconds."""
        super().__init__()
        self._lock = _lock = Lock()
        self._urls = _urls = dict()
        self._max_age = max_age

        def scanner():
            nonlocal _urls, _lock
            client = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            # see https://stackoverflow.com/questions/14388706
            client.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
            client.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            client.bind(("", Secrets.get_attr('broadcast_port', 50000)))
            try:
                while True:
                    data, addr = client.recvfrom(256)
                    try:
                        url, uid = data.split(b'\n')
                        # TODO: discard devices without device_config
                        with _lock:
                            _urls[url.decode()] = time.monotonic()
                    except (AttributeError, ValueError):
                        logger.DEBUG(f"received malformed advertisement {data} from {addr}")
            finally:
                client.close()


        self._thread = Thread(target=scanner)
        self._thread.start()

    def scan(self):
        with self._lock:
            # set of urls seen in last _max_age seconds
            return { url for url, t in self._urls.items() if (time.monotonic()-t) < self._max_age }


# device_conifg.py

from .env import Env
from .utilities import cd
from glob import glob
from fnmatch import fnmatch
import yaml, os

"""
DeviceConfig (from yaml): name, uid, resources (rsync)

Sample yaml description:

robot-stm:
    uid: 2d:00:49:00:09:50:52:42:4e:30:39:20
    install-dir: /spi/lib
    include-patterns: 
        - "*.py"
        - "*.mpy"
    exclude-patterns:
        - "boot_out.txt"
        - "/data"
    path: $IOT_PROJECTS
    resources:
        - pystone.py: /flash
        - copy.py
        - abc.py:
        - boot: 
            path: my_project/code
            unpack: true
            install-dir: /flash
            include-patterns:
                - "*.py"
        - boot: 
            lib: ~/my_own_library
            unpack: false
            install-dir: /flash

"""

class DeviceConfig:
    
    def __init__(self, name, uid, spec, file):
        """Not usually called directly. Use get_device_config(s)"""
        self._name = name
        self._uid = uid
        self._spec = spec
        self._file = file
    
    @property
    def name(self):
        return self._name
    
    @property
    def uid(self):
        return self._uid

    @property
    def file(self):
        return self._file

    def __str__(self):
        from io import StringIO
        s = StringIO()
        s.write(f"Configuration (in {self.file}):\n")
        for r in self.resources:
            s.write(f"  {r}\n")
        # s.write(f"    spec:                 {self._spec}\n")
        return s.getvalue()

    @staticmethod
    def get_device_config(name_or_uid):
        """Return DeviceConfig for device with given name or uid
        Raises ValueError if device not found.
        """
        devs = DeviceConfig.get_device_configs()
        # check for name
        if devs.get(name_or_uid): return devs.get(name_or_uid)
        # search for uid
        for dev in devs.values():
            if dev.uid == name_or_uid: return dev
        raise ValueError(f"No configuration found for: '{name_or_uid}'")

    @staticmethod
    def get_device_configs():
        """Return dict name --> DeviceConfig"""
        result = {}
        names = set()
        uids  = set()
        try:
            with cd(Env.iot_devices()):
                for file in glob("*.yaml") + glob("*.yml"):
                    with open(file) as f:
                        for name, spec in yaml.safe_load(f.read()).items():
                            if name in names:
                                raise ValueError(f"File {file}: device '{name}' redefined")
                            names.add(name)
                            uid = spec.get('uid')
                            if not uid:
                                raise ValueError(f"File {file} device '{name}': field 'uid' is mandatory")
                            uids.add(uid)
                            result[name] = DeviceConfig(name, uid, spec, file)
        except FileNotFoundError as e:
            # folder Env.iot_devices() does not exist
            pass
        return result

    @property
    def resources(self):
        return [ _Resource(self, r) for r in self._spec.get('resources', []) ]
    
    @property
    def resource_files(self):
        """Returns a dict
           path_on_mcu -> (mtime, size, path_on_host)
        """
        result = {}
        sep = os.path.sep
        for r in self.resources:
            for f in r.files:
                mcu_file = sep.join(f.strip(sep).split(sep)[1:]) if r.unpack else f
                mcu_path = os.path.join(r.install_dir, mcu_file)
                host_path = Env.expand_path(os.path.join(r.path, f))
                # add folders so rsync won't delete them
                p = mcu_path
                while p != '/':
                    p = os.path.dirname(p)
                    result[os.path.normpath(p)] = (0, -1, '')
                # add the file
                result[mcu_path] = (
                    os.path.getmtime(host_path), 
                    os.path.getsize(host_path), 
                    host_path
                )            
        return result

class _Resource:
    """Single Resource specified in yaml file"""

    def __init__(self, dev, spec):
        self._dev = dev
        if isinstance(spec, str):
            self._resource = spec
            self._param = {}
        elif isinstance(spec, dict):
            self._resource = next(iter(spec.keys()))
            self._param = spec[self._resource]
            if not self._param:
                self._param = {}
            elif isinstance(self._param, str):
                self._param = { 'install-dir': self._param }
        else:
            # should never happen
            raise ValueError(f"Resource: expected dict, got {type(spec)}")

    @property
    def name(self):
        """Resource name, also file or directory name"""
        return self._resource
    
    @property
    def files(self):
        """List of files in this resource, path relative lib"""
        result = []
        includes = self._param.get('include-patterns', self._dev._spec.get('include-patterns', [ './**/*.py', './**/*.mpy', './**/' ]))
        excludes = self._param.get('exclude-patterns', self._dev._spec.get('exclude-patterns', [ 'boot_out.txt' ]))
        if isinstance(includes, str): includes = [ includes ]
        if isinstance(excludes, str): excludes = [ excludes ]
        path = os.path.join(self.path, self.name)
        if os.path.isfile(Env.expand_path(path)): return [ self.name ]
        try:
            with cd(path):
                for inc in includes:
                    for file in glob(inc, recursive=True):
                        if not os.path.isfile(file): continue
                        for ex in excludes:
                            if fnmatch(file, ex): continue
                        result.append(os.path.normpath(os.path.join(self.name, file)))
        except OSError:
            pass
        return result

    @property
    def unpack(self):
        """Upload directory (unpack False) or contents (unpack True)"""
        if 'unpack' in self._param:
            return self._param['unpack'] 
        full_path = os.path.join(self.path, self.name)
        return os.path.isdir(full_path) and not os.path.isfile(os.path.join(full_path, '__init__.py'))

    @property
    def install_dir(self):
        """Directory on mcu in which this resource is located"""
        d = self._param.get('install-dir', self._dev._spec.get('install-dir', '/'))
        return d if d.startswith('/') else '/' + d

    @property
    def path(self):
        """Library (folder) where this resource is located on the host."""
        p = self._param.get('path', self._dev._spec.get('path', Env.iot_projects()))
        p = Env.expand_path(p)
        if not os.path.isabs(p):
            p = os.path.join(Env.iot_projects(), p)
            p = Env.expand_path(p)
        return p

    def __str__(self):
        # return f"{self.lib}/{self.name} -> {self.install_dir}"
        return f"Res {self.name:22} install-dir={self.install_dir:22} path={self.path}"
    


# env.py

from .utilities import cd
import os

"""
Environment Variables

  Name          Default
* IOT_PROJECTS  ~/projects
* IOT_DEVICES   $IOT_PROJECTS/devices
* IOT_LIBS      $IOT_PROJECTS/libs
* IOT_SECRETS   $IOT_LIBS/secrets.py
"""

class Env:

    @staticmethod
    def expand_path(path):
        return os.path.expanduser(os.path.expandvars(path))

    @staticmethod
    def iot_projects():
        return os.getenv('IOT_PROJECTS', '~/projects')

    @staticmethod
    def iot_devices():
        return os.getenv('IOT_DEVICES', os.path.join(Env.iot_projects(), 'devices'))

    @staticmethod
    def iot_libs():
        return os.getenv('IOT_LIBS', os.path.join(Env.iot_projects(), 'libs'))
        
    @staticmethod
    def iot_secrets():
        return os.getenv('IOT_SECRETS', os.path.join(Env.iot_libs(), 'secrets.py'))
    
    @staticmethod
    def print_config():
        print("IOT_PROJECTS:", Env.iot_projects())
        print("IOT_DEVICES: ", Env.iot_devices())
        print("IOT_LIBS:    ", Env.iot_libs())
        print("IOT_SECRETS: ", Env.iot_secrets())

# ensure this is defined ...

if not os.getenv('IOT_PROJECTS'):
    os.environ['IOT_PROJECTS'] = '~/projects'
